@@ -15,21 +15,23 @@ var symbolQuery = flag.String("symbol_query", "test", "symbol query to use in be
 // BenchmarkWorkspaceSymbols benchmarks the time to execute a workspace symbols
 // request (controlled by the -symbol_query flag).
 func BenchmarkWorkspaceSymbols(b *testing.B) {
-	env := benchmarkEnv(b)
+	for name := range repos {
+		b.Run(name, func(b *testing.B) {
+			env := getRepo(b, name).sharedEnv(b)
+			symbols := env.Symbol(*symbolQuery) // warm the cache
 
-	// Make an initial symbol query to warm the cache.
-	symbols := env.Symbol(*symbolQuery)
+			if testing.Verbose() {
+				fmt.Println("Results:")
+				for i, symbol := range symbols {
+					fmt.Printf("\t%d. %s (%s)\n", i, symbol.Name, symbol.ContainerName)
+				}
+			}
 
-	if testing.Verbose() {
-		fmt.Println("Results:")
-		for i := 0; i < len(symbols); i++ {
-			fmt.Printf("\t%d. %s (%s)\n", i, symbols[i].Name, symbols[i].ContainerName)
-		}
-	}
+			b.ResetTimer()
 
-	b.ResetTimer()
-
-	for i := 0; i < b.N; i++ {
-		env.Symbol(*symbolQuery)
+			for i := 0; i < b.N; i++ {
+				env.Symbol(*symbolQuery)
+			}
+		})
 	}
 }
