@@ -8,14 +8,15 @@ import (
 	"fmt"
 	"strings"
 	"testing"
+	"time"
 
 	"github.com/google/go-cmp/cmp"
 	"golang.org/x/tools/gopls/internal/bug"
 	"golang.org/x/tools/gopls/internal/hooks"
+	"golang.org/x/tools/gopls/internal/lsp/fake"
+	"golang.org/x/tools/gopls/internal/lsp/protocol"
 	. "golang.org/x/tools/gopls/internal/lsp/regtest"
 	"golang.org/x/tools/internal/testenv"
-
-	"golang.org/x/tools/gopls/internal/lsp/protocol"
 )
 
 func TestMain(m *testing.M) {
@@ -527,13 +528,60 @@ func main() {
 		env.AcceptCompletion(loc, completions.Items[0])
 		env.Await(env.DoneWithChange())
 		got := env.BufferText("main.go")
-		want := "package main\r\n\r\nimport (\r\n\t\"fmt\"\r\n\t\"math\"\r\n)\r\n\r\nfunc main() {\r\n\tfmt.Println(\"a\")\r\n\tmath.Sqrt(${1:})\r\n}\r\n"
+		want := "package main\r\n\r\nimport (\r\n\t\"fmt\"\r\n\t\"math\"\r\n)\r\n\r\nfunc main() {\r\n\tfmt.Println(\"a\")\r\n\tmath.Sqrt(${1:x float64})\r\n}\r\n"
 		if diff := cmp.Diff(want, got); diff != "" {
 			t.Errorf("unimported completion (-want +got):\n%s", diff)
 		}
 	})
 }
 
+func TestUnimportedCompletionHasPlaceholders60269(t *testing.T) {
+	// We can't express this as a marker test because it doesn't support AcceptCompletion.
+	const src = `
+-- go.mod --
+module example.com
+go 1.12
+
+-- a/a.go --
+package a
+
+var _ = b.F
+
+-- b/b.go --
+package b
+
+func F0(a, b int, c float64) {}
+func F1(int, chan *string) {}
+`
+	WithOptions(
+		WindowsLineEndings(),
+	).Run(t, src, func(t *testing.T, env *Env) {
+		env.OpenFile("a/a.go")
+		env.Await(env.DoneWithOpen())
+
+		// The table lists the expected completions as they appear in Items.
+		const common = "package a\r\n\r\nimport \"example.com/b\"\r\n\r\nvar _ = "
+		for i, want := range []string{
+			common + "b.F0(${1:a int}, ${2:b int}, ${3:c float64})\r\n",
+			common + "b.F1(${1:_ int}, ${2:_ chan *string})\r\n",
+		} {
+			loc := env.RegexpSearch("a/a.go", "b.F()")
+			completions := env.Completion(loc)
+			if len(completions.Items) == 0 {
+				t.Fatalf("no completion items")
+			}
+			saved := env.BufferText("a/a.go")
+			env.AcceptCompletion(loc, completions.Items[i])
+			env.Await(env.DoneWithChange())
+			got := env.BufferText("a/a.go")
+			if diff := cmp.Diff(want, got); diff != "" {
+				t.Errorf("%d: unimported completion (-want +got):\n%s", i, diff)
+			}
+			env.SetBufferContent("a/a.go", saved) // restore
+		}
+	})
+}
+
 func TestPackageMemberCompletionAfterSyntaxError(t *testing.T) {
 	// This test documents the current broken behavior due to golang/go#58833.
 	const src = `
@@ -575,6 +623,66 @@ func main() {
 	})
 }
 
+func TestCompleteAllFields(t *testing.T) {
+	// This test verifies that completion results always include all struct fields.
+	// See golang/go#53992.
+
+	const src = `
+-- go.mod --
+module mod.com
+
+go 1.18
+
+-- p/p.go --
+package p
+
+import (
+	"fmt"
+
+	. "net/http"
+	. "runtime"
+	. "go/types"
+	. "go/parser"
+	. "go/ast"
+)
+
+type S struct {
+	a, b, c, d, e, f, g, h, i, j, k, l, m int
+	n, o, p, q, r, s, t, u, v, w, x, y, z int
+}
+
+func _() {
+	var s S
+	fmt.Println(s.)
+}
+`
+
+	WithOptions(Settings{
+		"completionBudget": "1ns", // must be non-zero as 0 => infinity
+	}).Run(t, src, func(t *testing.T, env *Env) {
+		wantFields := make(map[string]bool)
+		for c := 'a'; c <= 'z'; c++ {
+			wantFields[string(c)] = true
+		}
+
+		env.OpenFile("p/p.go")
+		// Make an arbitrary edit to ensure we're not hitting the cache.
+		env.EditBuffer("p/p.go", fake.NewEdit(0, 0, 0, 0, fmt.Sprintf("// current time: %v\n", time.Now())))
+		loc := env.RegexpSearch("p/p.go", `s\.()`)
+		completions := env.Completion(loc)
+		gotFields := make(map[string]bool)
+		for _, item := range completions.Items {
+			if item.Kind == protocol.FieldCompletion {
+				gotFields[item.Label] = true
+			}
+		}
+
+		if diff := cmp.Diff(wantFields, gotFields); diff != "" {
+			t.Errorf("Completion(...) returned mismatching fields (-want +got):\n%s", diff)
+		}
+	})
+}
+
 func TestDefinition(t *testing.T) {
 	testenv.NeedsGo1Point(t, 17) // in go1.16, The FieldList in func x is not empty
 	files := `
